<#--
Template for generating C code from a Statestep model.
Author: Michael Breen
See the associated files for licence terms and usage notes.
-->
/* ${pp.outputFileName} - generated from ${pp.sourceFileName} */

#include <stdio.h>
#include <string.h>

<#assign c_file = "${pp.outputFileName}"/>
<#assign h_file = c_file?split(".c")[0] + ".h"/>
#include "${h_file}"
<#--
Put the following declarations in the .h file.
(An alternative to this trick is of course to create
a separate template for the header file.)
-->
<@pp.changeOutputFile name=h_file/>
/* ${pp.outputFileName} - generated from ${pp.sourceFileName} */

/* ******************* Core Interface ******************* */

<#-- Disregard "phantom variables" (variables with no values) -->
<#assign usable_variables =
 doc.model.variables["variable[./value]"]
/>
/* Variables (with at least one value). */
typedef enum
{
<#list usable_variables as var>
   ${variable_id(var.@name)},
</#list>
   s2_NUM_VARS
} s2_Variable;

/* Variable values (bitfields). */
<#list usable_variables as var>

typedef enum
{
<#assign count = 1 />
   <#list var[".//value[not(./value)]"] as leaf>
   ${value_id(leaf)} = ${count},
      <#assign count = count * 2 />
   </#list>
   ${dont_care_id(var.@name)} = ${count - 1}
} s2_Values_${var.@name};
</#list>

/*
Unless otherwise stated, a combination must associate a
value with each variable, that is, exactly one bit must
be set in each element of the contained array.
(During internal processing only, instances of this type
may be "combination masks", that is, a variable may
have multiple bits set, indicating that it is allowed
to have any of a number of values.)
*/
typedef struct
{
   /*
   A bitset for each variable, with the bit which is
   set indicating that variable's value.
   */
   int values[s2_NUM_VARS];
} s2_Combination;

/* Events. */
typedef enum
{
<#list doc.model.events.event as event>
   s2_e_${event.@name},
</#list>
   s2_NUM_EVENTS
} s2_Event;

/*
A pointer to a function that implements any rule logic
outside the model. When s2_applyRules() finds a rule
that it thinks applies, it can call a function of this
type to check if some condition it doesn't know about
makes the rule inapplicable.
(See the HTML documentation included with the template
package for an example.)
Returns: 1 if the proposed rule is OK, 0 to veto it.
Parameters:
event: the event being responded to.
rule: a rule which seems to apply, given the event
   and the current combination
current: the current state
next: This is the result of applying the proposed rule
   to the current combination. In the case of an
   uncertain result (non-deterministic rule), one
   will have been chosen arbitrarily; this can be
   changed if appropriate.
*/
typedef int (*s2_ExtraRuleLogic)(s2_Event event, int rule,
   s2_Combination current, s2_Combination *next);

int s2_applyRules(s2_Event event, s2_Combination current,
   s2_Combination *next, s2_ExtraRuleLogic extraLogic);

int s2_stateAllowed(s2_Combination);

int s2_eventAllowed(s2_Event, s2_Combination);


/* ****** Extras Available for Test or Other Code ******* */

<#-- number of values of each variable -->
<#assign total_combs = 1/>
<#list usable_variables as var>
   <#assign total_combs = total_combs *
    var[".//value[not(./value)]"]?size
   />
</#list>
/* The number of possible combinations, ignoring constraints. */
#define s2_NUM_COMBINATIONS ${total_combs}

int s2_combinationToIndex(s2_Combination comb);

s2_Combination s2_combinationFromIndex(int index);

extern int s2_numValues[s2_NUM_VARS];

extern char* s2_variableNames[s2_NUM_VARS];

char* s2_getValueName(s2_Variable var, int value);

extern char* s2_eventNames[/*s2_NUM_EVENTS*/];

extern int s2_numRules[/*s2_NUM_EVENTS*/];

extern char** s2_ruleNames[/*s2_NUM_EVENTS*/];

/* The number of global constraint formulas. */
#define s2_NUM_GLOBAL_CONSTRAINTS <#--
-->${doc.model.constraints.constraint?size}

extern char* s2_globalConstraintNames[/*s2_NUM_GLOBAL_CONSTRAINTS*/];

int s2_constraintTableAllows(s2_Combination state);

/* Pointer to function implementing a constraint formula. */
typedef int (*s2_ConstraintFormula)(s2_Combination);

extern s2_ConstraintFormula
   s2_globalConstraints[/*s2_NUM_GLOBAL_CONSTRAINTS*/];

extern int s2_numEventConstraints[/*s2_NUM_EVENTS*/];

extern s2_ConstraintFormula* s2_eventConstraints[/*s2_NUM_EVENTS*/];

extern char** s2_constraintNames[/*s2_NUM_EVENTS*/];

int s2_ruleApplies(s2_Event event, int rule,
   s2_Combination current, s2_Combination *next);

<#-- switch back to the C file for the remaining output -->
<@pp.changeOutputFile name=c_file append=true/>

/*
NOTE ON ARRAY DIMENSIONS:
An extra padding element is added to some arrays, making
them one element larger than they should be.
This avoids the danger of any arrays being of size 0
(which are not strictly allowed in C) in a manner that
simplifies code generation.
Padding elements are not added to arrays of dimension
s2_NUM_VARS since, if the generated code is to be useful,
at least some variables must have been defined.
*/

/* ****** Types, Constants, Advance Declarations ******** */

extern char** s2_valueNames[s2_NUM_VARS];

/* Number of rows in the constraint table. */
#define s2_CONSTRAINT_TABLE_ROWS <#--
-->${doc.model.simple_constraints.constraint_group?size}

extern int
   s2_constraintTable[/*s2_CONSTRAINT_TABLE_ROWS*/][s2_NUM_VARS + 1];

/*
The tabular part of a rule, indexed by row and variable.
The source row (current variable values) is first, the
target row (new values of the variables) second.
Values of '0' on the target row of a rule indicate that the
corresponding variable's value does not change.
*/
typedef int s2_RuleTableCells[2][s2_NUM_VARS];

extern s2_RuleTableCells* s2_ruleTableParts[/*s2_NUM_EVENTS*/];

/* Pointer to function implementing a rule formula. */
typedef int (*s2_RuleFormula)(s2_Combination, s2_Combination *);

extern s2_RuleFormula* s2_ruleFormulas[/*s2_NUM_EVENTS*/];

int s2_globalConstraintsAllow(s2_Combination state);

void s2_collapseToCombination(s2_Combination *comb);

s2_Combination s2_getDontCares();

/* ****************** Core Functions ******************** */

/*
If any rule applies in the current state for the specified
event then set the next state according to the first such
rule and return (the index of) the rule.
Otherwise return -1.
If the extraLogic parameter is not null then the function
pointed to will be given a chance to confirm the
applicability of any rule that otherwise seems to apply.
See the s2_ExtraRuleLogic type definition and, for an
example, the HTML documentation included with the template.
Note that if the rule allows for more than one next state
for a given current state (e.g., if the value to which some
variable changes is "don't care" according to the rule)
then one possible next state is chosen arbitrarily.
This function does not check for consistency with the
constraints.
*/
int s2_applyRules(s2_Event event, s2_Combination current,
   s2_Combination *next, s2_ExtraRuleLogic extraLogic)
{
   s2_Combination nextTemp;
   int rule;
   for (rule = 0; rule < s2_numRules[event]; ++rule)
   {
      nextTemp = s2_getDontCares();
      if (s2_ruleApplies(event, rule, current, &nextTemp))
      {
         if (extraLogic)
         {
            if (!(*extraLogic)(event, rule, current, &nextTemp))
               continue;
         }
         *next = nextTemp;
         return rule;
      }
   }
   return -1;
}

/*
Returns 1 if the constraint table and the global constraint
formulas allow the specified combination, otherwise 0.
*/
int s2_stateAllowed(s2_Combination state)
{
   return s2_constraintTableAllows(state) &&
      s2_globalConstraintsAllow(state);
}

/*
Returns 1 if the constraint formulas for the specified event
allow the specified combination, otherwise 0.
*/
int s2_eventAllowed(s2_Event event, s2_Combination state)
{
   int numCons = s2_numEventConstraints[event];
   s2_ConstraintFormula *cons = s2_eventConstraints[event];
   int i;
   for (i = 0; i < numCons; ++i)
   {
      if (!(*cons[i])(state))
         return 0;
   }
   return 1;
}

/* ***************** Extra Functions ******************** */

/*
Returns the name of the specified value.
Returns 0 (NULL) if the value does not correspond
to any single value of the specified variable.
*/
char* s2_getValueName(s2_Variable var, int value)
{
   int index;
   for (index = 0; value; ++index)
   {
      if (value & 1)
      {
         if (value & ~1)
            return 0;
         return s2_valueNames[var][index];
      }
      value >>= 1;
   }
   return 0;
}

/*
Maps every combination to a unique number between 0 and
s2_NUM_COMBINATIONS - 1.
*/
int s2_combinationToIndex(s2_Combination comb)
{
   int result = 0;
   int factor = 1;
   int var;
   for (var = s2_NUM_VARS - 1; var >= 0; --var)
   {
      int offset = 0;
      int val;
      for (val = comb.values[var]; val & ~1; val >>= 1)
         ++offset;
      result += offset * factor;
      factor *= s2_numValues[var];
   }
   return result;
}

/*
The inverse of s2_combinationToIndex():
Maps each number between 0 and s2_NUM_COMBINATIONS - 1
to a unique combination.
(Larger numbers wrap-around: s2_NUM_COMBINATIONS maps
to the same combination as 0.)
*/
s2_Combination s2_combinationFromIndex(int index)
{
   s2_Combination comb;
   int var;
   for (var = s2_NUM_VARS - 1; var >= 0; --var)
   {
      int offset = index % s2_numValues[var];
      comb.values[var] = (1 << offset);
      index /= s2_numValues[var];
   }
   return comb;
}

/*
Returns 1 if the constraint table allows the specified
combination, otherwise 0.
*/
int s2_constraintTableAllows(s2_Combination state)
{
   int row;
   for (row = 0; row < s2_CONSTRAINT_TABLE_ROWS; ++row)
   {
      int ifVar = s2_constraintTable[row][s2_NUM_VARS];
      int ifVals = s2_constraintTable[row][ifVar];
      if (state.values[ifVar] & ifVals)
      {
         int thenVar;
         for (thenVar = 0; thenVar < s2_NUM_VARS; ++thenVar)
         {
            if (!(state.values[thenVar] &
               s2_constraintTable[row][thenVar]))
            {
               return 0;
            }
         }
      }
   }
   return 1;
}

/*
Check if the specified rule for the specified event
allows a transition from the specified current (source)
state to the specified next (target) state.
The 'next' parameter may be a combination mask, i.e.,
multiple values (e.g., "don't care") may be set for a
variable to indicate it is allowed any of those values.
In this case, if the rule applies then 'next' will be
changed to exactly one of the states allowed by the
combination mask that is consistent with the rule.
Return 1 if the rule applies, otherwise 0.
*/
int s2_ruleApplies(s2_Event event, int rule,
   s2_Combination current, s2_Combination *next)
{
   s2_RuleFormula ruleFormula;
   s2_Combination nextTemp = *next;
   int variable;
   for (variable = 0; variable < s2_NUM_VARS; ++variable)
   {
      int sMask, tMask;
      sMask = s2_ruleTableParts[event][rule][0][variable];
      tMask = s2_ruleTableParts[event][rule][1][variable];
      nextTemp.values[variable] &= (tMask ? tMask :
         current.values[variable]);
      if (!(sMask & current.values[variable])
         || !nextTemp.values[variable])
      {
         return 0;
      }
   }
   ruleFormula = s2_ruleFormulas[event][rule];
   if (ruleFormula)
   {
      if (!(*ruleFormula)(current, &nextTemp))
         return 0;
   }
   s2_collapseToCombination(&nextTemp);
   *next = nextTemp;
   return 1;
}

/* ***************** Local Functions ******************** */

/*
Returns 1 if the global constraint formulas allow the specified
combination, otherwise 0.
*/
int s2_globalConstraintsAllow(s2_Combination state)
{
   int con;
   for (con = 0; con < s2_NUM_GLOBAL_CONSTRAINTS; ++con)
   {
      if (!(*s2_globalConstraints[con])(state))
         return 0;
   }
   return 1;
}

/*
Changes a combination mask, that is, a combination in which
a variable is allowed to have any of one or more values,
into a combination in which every variable has one of the
values allowed (chosen arbitrarily).
*/
void s2_collapseToCombination(s2_Combination *comb)
{
   int var;
   for (var = 0; var < s2_NUM_VARS; ++var)
   {
      int valSet = comb->values[var];
      int val = 1;
      while (val && !(val & valSet))
         val <<= 1;
      comb->values[var] = val;
   }
}

/*
Returns a combination mask in which every variable is
"don't care".
*/
s2_Combination s2_getDontCares()
{
   s2_Combination comb;
<#list usable_variables as var>
   comb.values[${variable_id(var.@name)}] = <#--
      -->${dont_care_id(var.@name)};
</#list>
   return comb;
}

/* *************** Variable Definitions ***************** */

/* Variable names. */
char* s2_variableNames[s2_NUM_VARS] = {
<#list usable_variables as var>
   "${var.@name}",
</#list>
   };

/* Variable value names. */
<#list usable_variables as var>

char* s2_valueNames_${var.@name}[] = {
   <#list var[".//value[not(./value)]"] as leaf>
   "<#list leaf?ancestors("value")?reverse as branch>${branch.@name}-</#list
   >${leaf.@name}"<#if leaf_has_next>,</#if>
   </#list>
   };
</#list>

/* Array of all value names indexed by variable. */
char** s2_valueNames[s2_NUM_VARS] = {
<#list usable_variables as var>
   s2_valueNames_${var.@name}<#if var_has_next>,</#if>
</#list>
   };

/* Array of "don't care" values indexed by variable. */
int s2_dontCares[s2_NUM_VARS] = {
<#list usable_variables as var>
   ${dont_care_id(var.@name)}<#if var_has_next>,</#if>
</#list>
   };

/* Number of values defined for each variable. */
int s2_numValues[s2_NUM_VARS] = {
<#list usable_variables as var>
   ${var[".//value[not(./value)]"]?size}<#if var_has_next>,</#if>
</#list>
   };

/* Event names. */
char* s2_eventNames[/*s2_NUM_EVENTS*/] = {
<#list doc.model.events.event as event>
   "${event.@name}",
</#list>
   "" /* padding */
   };

/* The number of rules for each event, indexed by event. */
int s2_numRules[/*s2_NUM_EVENTS*/] = {
<#list doc.model.events.event as event>
   ${event.rules.rule?size},
</#list>
   0 /* padding */
   };

/* Rule names. */
<#list doc.model.events.event as event>

char* s2_ruleNames_${event.@name}[] = {
   <#list event.rules.rule as rule>
   "${rule.@name}",
   </#list>
   "" /* padding */
   };
</#list>

/* Rule names, indexed by [event][rule]. */
char** s2_ruleNames[/*s2_NUM_EVENTS*/] = {
<#list doc.model.events.event as event>
   s2_ruleNames_${event.@name},
</#list>
   0 /* padding */
   };

/*
Constraint table indexed by [row][variable].
The extra element at the end of each row indicates the variable
(column index) containing the constraining values for that row.
*/
<#if doc.model.simple_constraints.constraint_group?size == 0>
int s2_constraintTable[1][s2_NUM_VARS + 1]; /* empty */
<#else>
int s2_constraintTable[s2_CONSTRAINT_TABLE_ROWS][s2_NUM_VARS + 1]
   = {
   <#list doc.model.simple_constraints.constraint_group as row>
   {
<@values_array value_sets=row.value_set trailing=","/>
   ${variable_id(row["./value_set[./@type='constraining']/@variable"][0])} /* constraining variable */
   }<#if row_has_next>,</#if>
   </#list>
   };
</#if>

/*
Constraint functions return 1 if the specified combination
(state) satisfies the corresponding constraint, otherwise 0.

Note that it is assumed the argument is a single combination,
that is, exactly one value is associated with each variable.
*/

<#list doc.model.constraints.constraint as constraint>
   <#assign formula = constraint.formula/>
/* Global constraint '${constraint.@name}':
${formula.text?replace("*/", "* /")?replace("/*", "/ *")}
*/
int s2_c_${constraint.@name}(s2_Combination current)
{
   <#if formula.parse_node?size != 0>
      <@parse_node_to_code node=formula.parse_node />
   </#if>
   return 1;
}

</#list>
/* Global constraint formulas. */
s2_ConstraintFormula
   s2_globalConstraints[/*s2_NUM_GLOBAL_CONSTRAINTS*/] = {
<#list doc.model.constraints.constraint as constraint>
   &s2_c_${constraint.@name},
</#list>
   0 /* padding */
   };


/* Global constraint names. */
char* s2_globalConstraintNames[/*s2_NUM_GLOBAL_CONSTRAINTS*/] = {
<#list doc.model.constraints.constraint as constraint>
   "${constraint.@name}",
</#list>
   "" /* padding */
   };


<#list doc.model.events.event as event>
   <#list event.constraints.constraint as constraint>
      <#assign formula = constraint.formula/>
/* Event '${event.@name}', constraint '${constraint.@name}':
${formula.text?replace("*/", "* /")}
*/
int s2_c_${event.@name}__${constraint.@name}(s2_Combination current)
{
      <#if formula.parse_node?size != 0>
         <@parse_node_to_code node=formula.parse_node />
      </#if>
   return 1;
}

   </#list>
/* Constraint formulas for event '${event.@name}'. */
s2_ConstraintFormula s2_constraints_${event.@name}[] = {
   <#list event.constraints.constraint as constraint>
   &s2_c_${event.@name}__${constraint.@name},
   </#list>
   0 /* padding */
   };

</#list>

/* All of the event constraints, indexed by [event][constraint]. */
s2_ConstraintFormula* s2_eventConstraints[/*s2_NUM_EVENTS*/] = {
<#list doc.model.events.event as event>
   s2_constraints_${event.@name},
</#list>
   0 /* padding */
   };

/* The number of constraints for each event. */
int s2_numEventConstraints[/*s2_NUM_EVENTS*/] = {
<#list doc.model.events.event as event>
   ${event.constraints.constraint?size},
</#list>
   0 /* padding */
   };

<#list doc.model.events.event as event>
/* Constraint names for event '${event.@name}'. */
char* s2_constraintNames_${event.@name}[] = {
   <#list event.constraints.constraint as constraint>
   "${constraint.@name}",
   </#list>
   "" /* padding */
   };

</#list>

/* All the event constraint names, indexed by [event][constraint]. */
char** s2_constraintNames[/*s2_NUM_EVENTS*/] = {
<#list doc.model.events.event as event>
   s2_constraintNames_${event.@name},
</#list>
   0 /* padding */
   };

/* The tabular parts of the rules for each of the events. */
<#list doc.model.events.event as event>

   <#if event.rules.rule?size != 0>
s2_RuleTableCells s2_ruleTableParts_${event.@name}[] = {
      <#list event.rules.rule as rule>
   /* Rule '${rule.@name}': */
   { {
<@values_array
 value_sets=rule["./values_array/value_set[@type = 'source']"]
/>
   }, {
<@values_array
 value_sets=rule["./values_array/value_set[@type = 'target']"]
/>
   } }<#if rule_has_next>,</#if>
      </#list>
   };
   </#if>
</#list>

/* The tabular parts of rules, indexed by [event][rule]. */
s2_RuleTableCells* s2_ruleTableParts[/*s2_NUM_EVENTS*/] = {
<#list doc.model.events.event as event>
   <#if event.rules.rule?size != 0>
   s2_ruleTableParts_${event.@name},
   <#else>
   0,
   </#if>
</#list>
   0 /* padding */
   };

/*
Functions for rule formulas.

Parameter 'current' (current state):
Each variable must be associated with exactly one value,
i.e., a single bit must be set for each.

Parameter 'next' (next state):
This may be a combination mask.
Where variables are associated with a value or a set of
values other than "don't care", this acts to constrain the
ways in which the formula can be satisfied. For example,
if *next has the value x1 for some variable X and the
formula is "X->{x2} || (X->{=} && Y->{y3}" then the
formula will be satisfied only if the current value of X
is also x1 and *next allows Y to be y3.
If the formula is satisfied then *next is changed where
necessary to be consistent with the formula - so, in the
example above, Y might be changed from "don't care" to y3.
However, other variables might still be allowed several
values, i.e., *next may still be a combination mask after
the function returns.

Functions return 1 if the formula is satisfied; otherwise
0 is returned and *next is unchanged.
*/

<#list doc.model.events.event as event>
   <#list event.rules.rule.formula as formula>
      <#assign rule = formula?parent/>
      <#assign function_id =
       "s2_r_${event.@name}__${rule.@name}"
      />
/* Event '${event.@name}', rule '${rule.@name}':
${formula.text?replace("*/", "* /")}
*/
int ${function_id}(s2_Combination current,
   s2_Combination *next)
{
      <#if formula.parse_node?size != 0>
         <#-- if there are any assignments in the formula -->
         <#if formula[".//value_set[@type='target']"]?size != 0>
            <#-- then we'll be using these variables -->
   int mask, restricted, excluded;
         </#if>
         <@parse_node_to_code node=formula.parse_node />
      </#if>
   return 1;
}

   </#list>
/* Rule formulas for event '${event.@name}', indexed by rule. */
s2_RuleFormula s2_ruleFormulas_${event.@name}[] = {
   <#list event.rules.rule as rule>
      <#if rule.formula?size != 0>
   &s2_r_${event.@name}__${rule.@name},
      <#else>
   0,
      </#if>
   </#list>
   0 /* padding */
   };

</#list>

/*
Rule formulas, indexed by [event][rule];
0 (NULL) wherever a rule has no formula.
*/
s2_RuleFormula* s2_ruleFormulas[/*s2_NUM_EVENTS*/] = {
<#list doc.model.events.event as event>
   s2_ruleFormulas_${event.@name},
</#list>
   0 /* padding */
   };

<#----------------- Macros and functions ------------------>
<#--
This recursive macro converts a parse tree for a constraint
formula or a rule formula into code.
While the generated code will be somewhat inefficient in
some cases (e.g., for a formula like "MyVar -> {=}"), it
must be remembered that the approach used is designed to
ensure more complex formulas like
"A->{a1} && B={b1} || !A->{=} && B={b2|b3}"
are transformed correctly.
This is difficult because formulas use a declarative syntax
which must be mapped to imperative program code.
In particular, for changes to the value of a variable in a
rule formula ("Var -> ..."), some kind of lookahead or
backtracking is needed in case it is later found that the
assignment should not have been made. For this, the
generated code uses recursion.
Simplicity and robustness in the macro are favoured over
elegance or efficiency of the generated code.
-->
<#macro parse_node_to_code node fail_cmd="return 0;" indent="   ">
   <#if node.@type == "values">
      <#local vals = node.value_set/>
      <#local var_id = variable_id(vals.@variable)/>
      <#if vals.@type == "source">
${indent}if (!(current.values[${var_id}] & (<@values_to_code vals/>)))
${indent}   ${fail_cmd}
      <#elseif vals.@type == "target">
         <#if vals.@content == "no_change">
${indent}mask = current.values[${var_id}];
         <#else>
${indent}mask = <@values_to_code vals/>;
         </#if>
${indent}restricted = next->values[${var_id}] & mask;
${indent}excluded = next->values[${var_id}] & ~mask;
${indent}if (!restricted)
${indent}{
${indent}   ${fail_cmd}
${indent}}
${indent}else if (excluded)
${indent}{
${indent}   s2_Combination nextCopy = *next;
${indent}   nextCopy.values[${var_id}] = restricted;
${indent}   if (${function_id}(current, &nextCopy))
${indent}   {
${indent}      *next = nextCopy;
${indent}      return 1;
${indent}   }
${indent}   nextCopy.values[${var_id}] = excluded;
${indent}   if (${function_id}(current, &nextCopy))
${indent}   {
${indent}      *next = nextCopy;
${indent}      return 1;
${indent}   }
${indent}   return 0;
${indent}}
      <#else>
ERROR: Unexpected value_set type ("${vals.@type}").
      </#if>
   <#elseif node.@type == "not">
      <#local var_id = "inverse_${uniq()}"/>
${indent}int ${var_id} = 1;
      <@parse_node_to_code node=node.parse_node
       fail_cmd="${var_id} = 0;" indent=indent
      />
${indent}if (${var_id})
${indent}   ${fail_cmd}
   <#elseif node.@type == "and">
      <#local var_id = "conjunct_${uniq()}"/>
      <#list node.parse_node as subnode>
         <#if subnode_index == 0>
${indent}int ${var_id} = 1;
            <@parse_node_to_code node=subnode
             fail_cmd = "${var_id} = 0;" indent=indent
            />
         <#else>
${indent}if (${var_id})
${indent}{
            <@parse_node_to_code node=subnode
             fail_cmd = "${var_id} = 0;" indent=indent+"   "
            />
${indent}}
         </#if>
      </#list>
${indent}if (!${var_id})
${indent}   ${fail_cmd}
   <#elseif node.@type == "or">
      <#local var_id = "disjunct_${uniq()}"/>
      <#list node.parse_node as subnode>
         <#if subnode_index == 0>
${indent}int ${var_id} = 1;
            <@parse_node_to_code node=subnode
             fail_cmd = "${var_id} = 0;" indent=indent
            />
         <#else>
${indent}if (!${var_id})
${indent}{
${indent}   ${var_id} = 1;
${indent}   {
            <@parse_node_to_code node=subnode
             fail_cmd = "${var_id} = 0;" indent=indent+"      "
            />
${indent}   }
${indent}}
         </#if>
      </#list>
${indent}if (!${var_id})
${indent}   ${fail_cmd}
   <#elseif node.@type == "implies">
      <#local var_id = "antecedent_${uniq()}"/>
${indent}int ${var_id} = 1;
      <@parse_node_to_code node=node.parse_node[0]
       fail_cmd="${var_id} = 0;" indent=indent
      />
${indent}if (${var_id})
${indent}{
      <@parse_node_to_code node=node.parse_node[1]
       fail_cmd=fail_cmd indent=indent+"   "
      />
${indent}}
   <#else>
ERROR: Unexpected parse_node type ("${node.@type}").
   </#if>
</#macro>

<#--
A function that returns the number of times it has been
called (this number can be appended to identifiers to
ensure uniqueness).
-->
<#function uniq>
   <#assign uniq_count = (uniq_count + 1)?default(1)/>
   <#return uniq_count/>
</#function>

<#--
Take zero or more 'value_set' nodes and output a list of
variable values, one set of values for each defined variable,
in the order in which the variables were originally defined;
for any variable for which there is no 'value_set', output a
"don't care".
-->
<#macro values_array value_sets indent_space="   " trailing="">
   <#list usable_variables as var>
      <#local found = false/>
      <#list value_sets as value_set>
         <#if value_set.@variable == '${var.@name}'>
            <#local found = true/>
${indent_space}<@values_to_code value_set/><#if
var_has_next>,<#else>${trailing}</#if>
            <#break/>
         </#if>
      </#list>
      <#if !found>
${indent_space}${dont_care_id(var.@name)}<#if
var_has_next>,<#else>${trailing}</#if>
      </#if>
   </#list>
</#macro>

<#--
Convert a set of variable values to the corresponding code.
-->
<#macro values_to_code valset>
   <#switch valset.@content>
<#case "dont_care">${dont_care_id(valset.@variable)}<#break>
<#case "no_change">0<#break>
      <#case "list">
         <#local children=valset.value/>
         <#if children?size == 0>
            <#-- an empty list is an implicit "don't care" -->
${dont_care_id(valset.@variable)}<#--
-->
         <#else>
            <#list val_def_nodes(valset) as val>
${value_id(val)}<#if val_has_next> | </#if
>
            </#list>
         </#if>
      <#break>
   </#switch>
</#macro>

<#--
Given a set of (leaf) value nodes in a value set,
return the corresponding leaf nodes in the variable
value definition tree.
The trick here is that one or more input nodes may
actually be a group of values, i.e., it may not
correspond to a leaf node in the definition tree.
In this case, the leaf nodes below the corresponding
definition node will be returned, giving a longer
sequence than was passed in.
-->
<#function val_def_nodes value_set>
   <#local result = []/>
   <#local varName = value_set.@variable/>
   <#local var =
    doc.model.variables["variable[@name='${varName}']"][0]
   />
   <#list value_set[".//value[not(./value)]"] as group>
      <#local def_node = var/>
      <#local path_to_group = group?ancestors("value")?reverse
       + [group]/>
      <#list path_to_group as component>
         <#local name=component.@name/>
         <#local def_node = def_node["./value[@name='${name}']"][0]/>
      </#list>
      <#local leaf_list = def_node[".//value[not(./value)]"]/>
      <#if leaf_list?size == 0>
         <#local leaf_list = [def_node]/>
      </#if>
      <#local result = result + leaf_list/>
   </#list>
   <#return result>
</#function>

<#--
Given a (leaf) value node in a variable definition tree,
i.e., below a 'variable' element, produce the
corresponding variable value identifier for the code.
-->
<#function value_id def_node>
   <#local id = def_node.@name/>
   <#list def_node?ancestors("value") as group>
      <#local id = group.@name + "_" + id/>
   </#list>
   <#local var_name = def_node?ancestors("variable")[0].@name/>
   <#local id = variable_id(var_name) + "__" + id/>
   <#return id/>
</#function>

<#--
Given the name of a variable in the model (a string),
return the variable identifier in the generated code.
-->
<#function variable_id var_name>
   <#return "s2_v_" + var_name/>
</#function>

<#--
Given the name of a variable in the model (a string),
return its "don't care" value in the generated code.
-->
<#function dont_care_id var_name>
   <#return variable_id(var_name)+ "_ALLVALS"/>
</#function>
